@@ -4,9 +4,8 @@
 //-----------------------------------------------------------------------------
 
 #define PY_SSIZE_T_CLEAN
-
 #include <Python.h>
-#ifdef MS_WINDOWS
+
 #define WIN32_LEAN_AND_MEAN
 #include <windows.h>
 #include <imagehlp.h>
@@ -49,17 +48,13 @@ typedef struct {
     WORD idCount;                       // How many images?
     GRPICONDIRENTRY idEntries[0];       // An entry for each image
 } GRPICONDIR;
-#endif
 
 //-----------------------------------------------------------------------------
 // Globals
 //-----------------------------------------------------------------------------
-#ifdef MS_WINDOWS
 static PyObject *g_BindErrorException = NULL;
 static PyObject *g_ImageNames = NULL;
-#endif
 
-#ifdef MS_WINDOWS
 //-----------------------------------------------------------------------------
 // BindStatusRoutine()
 //   Called by BindImageEx() at various points. This is used to determine the
@@ -102,31 +97,36 @@ static BOOL __stdcall BindStatusRoutine(
 //   Return the data for the given file.
 //-----------------------------------------------------------------------------
 static int GetFileData(
-    const char *fileName,               // name of file to read
+    wchar_t *filename,                  // name of file to read
     char **data)                        // pointer to data (OUT)
 {
-    DWORD numberOfBytesRead, dataSize;
-    HANDLE file;
+    DWORD bytesread, filesize;
+    HANDLE fhandle;
 
-    file = CreateFile(fileName, GENERIC_READ, FILE_SHARE_READ, NULL,
-            OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
-    if (file == INVALID_HANDLE_VALUE)
+    fhandle = CreateFileW((LPCWSTR) filename,
+                          GENERIC_READ,
+                          FILE_SHARE_READ,
+                          NULL,
+                          OPEN_EXISTING,
+                          FILE_ATTRIBUTE_NORMAL,
+                          NULL);
+    if (fhandle == INVALID_HANDLE_VALUE)
         return -1;
-    dataSize = GetFileSize(file, NULL);
-    if (dataSize == INVALID_FILE_SIZE) {
-        CloseHandle(file);
+    filesize = GetFileSize(fhandle, NULL);
+    if (filesize == INVALID_FILE_SIZE) {
+        CloseHandle(fhandle);
         return -1;
     }
-    *data = PyMem_Malloc(dataSize);
+    *data = PyMem_Malloc(filesize);
     if (!*data) {
-        CloseHandle(file);
+        CloseHandle(fhandle);
         return -1;
     }
-    if (!ReadFile(file, *data, dataSize, &numberOfBytesRead, NULL)) {
-        CloseHandle(file);
+    if (!ReadFile(fhandle, *data, filesize, &bytesread, NULL)) {
+        CloseHandle(fhandle);
         return -1;
     }
-    CloseHandle(file);
+    CloseHandle(fhandle);
     return 0;
 }
 
@@ -175,7 +175,9 @@ static PyObject *ExtAddIcon(
     PyObject *self,                     // passthrough argument
     PyObject *args)                     // arguments
 {
-    char *executableName, *iconName, *data, *iconData;
+    char *data, *iconData;
+    wchar_t *executable_name, *icon_name;
+    PyObject *executable, *icon;
     GRPICONDIR *groupIconDir;
     DWORD resourceSize;
     ICONDIR *iconDir;
@@ -183,30 +185,53 @@ static PyObject *ExtAddIcon(
     HANDLE handle;
     int i;
 
-    if (!PyArg_ParseTuple(args, "ss", &executableName, &iconName))
-        return NULL;
+    succeeded = TRUE;
+    handle = NULL;
+    data = NULL;
+    groupIconDir = NULL;
+    executable_name = NULL;
+    icon_name = NULL;
+
+    if (!PyArg_ParseTuple(args, "UU", &executable, &icon)) {
+        succeeded = FALSE;
+        PyErr_Format(
+            PyExc_RuntimeError, "Invalid parameters."
+        );
+    }
+
+    if (succeeded) {
+        executable_name = PyUnicode_AsWideCharString(executable, NULL);
+        icon_name = PyUnicode_AsWideCharString(icon, NULL);
+        if (!executable_name || !icon_name) {
+            succeeded = FALSE;
+            PyErr_NoMemory();
+        }
+    }
 
     // begin updating the executable
-    handle = BeginUpdateResource(executableName, FALSE);
-    if (!handle) {
-        PyErr_SetExcFromWindowsErrWithFilename(PyExc_WindowsError,
-                GetLastError(), executableName);
-        return NULL;
+    if (succeeded) {
+        handle = BeginUpdateResourceW(executable_name, FALSE);
+        if (!handle) {
+            succeeded = FALSE;
+            PyErr_SetExcFromWindowsErrWithFilenameObject(
+                PyExc_WindowsError, GetLastError(), executable
+            );
+        }
     }
 
     // first attempt to get the data from the icon file
-    data = NULL;
-    succeeded = TRUE;
-    groupIconDir = NULL;
-    if (GetFileData(iconName, &data) < 0)
-        succeeded = FALSE;
+    if (succeeded) {
+        if (GetFileData(icon_name, &data) < 0)
+            succeeded = FALSE;
+    }
 
     // check for valid icon
     if (succeeded) {
-        iconDir = (ICONDIR*) data;
+        iconDir = (ICONDIR *) data;
         if (iconDir->idType != 1) {
-            PyErr_Format(PyExc_RuntimeError,
-                "Icon filename '%s' has invalid type.", iconName);
+            PyErr_Format(
+                PyExc_RuntimeError, "Icon filename '%S' has invalid type.", icon
+            );
             succeeded = FALSE;
         }
     }
@@ -227,7 +252,8 @@ static PyObject *ExtAddIcon(
         for (i = 0; i < iconDir->idCount; i++) {
             iconData = &data[iconDir->idEntries[i].dwImageOffset];
             resourceSize = iconDir->idEntries[i].dwBytesInRes;
-            succeeded = UpdateResource(handle, RT_ICON, MAKEINTRESOURCE(i + 1),
+            succeeded = UpdateResource(handle, RT_ICON,
+                    MAKEINTRESOURCE(i + 1),
                     MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL), iconData,
                     resourceSize);
             if (!succeeded)
@@ -236,15 +262,22 @@ static PyObject *ExtAddIcon(
     }
 
     // finish writing the resource (or discarding the changes upon an error)
-    if (!EndUpdateResource(handle, !succeeded)) {
-        if (succeeded) {
-            succeeded = FALSE;
-            PyErr_SetExcFromWindowsErrWithFilename(PyExc_WindowsError,
-                    GetLastError(), executableName);
+    if (handle) {
+        if (!EndUpdateResourceW(handle, !succeeded)) {
+            if (succeeded) {
+                succeeded = FALSE;
+                PyErr_SetExcFromWindowsErrWithFilenameObject(
+                    PyExc_WindowsError, GetLastError(), executable
+                );
+            }
         }
     }
 
     // clean up
+    if (executable_name)
+        PyMem_Free(executable_name);
+    if (icon_name)
+        PyMem_Free(icon_name);
     if (groupIconDir)
         PyMem_Free(groupIconDir);
     if (data)
@@ -278,7 +311,7 @@ static PyObject *ExtBeginUpdateResource(
                 GetLastError(), fileName);
         return NULL;
     }
-    return PyLong_FromLongLong(handle);
+    return PyLong_FromVoidPtr(handle);
 }
 
 
@@ -293,10 +326,13 @@ static PyObject *ExtUpdateResource(
     int resourceType, resourceId, resourceDataSize;
     char *resourceData;
     HANDLE handle;
+    PyObject *handle_obj;
 
-    if (!PyArg_ParseTuple(args, "iiis#", &handle, &resourceType, &resourceId,
-            &resourceData, &resourceDataSize))
+    if (!PyArg_ParseTuple(args, "Oiis#", &handle_obj,
+                          &resourceType, &resourceId,
+                          &resourceData, &resourceDataSize))
         return NULL;
+    handle = PyLong_AsVoidPtr(handle_obj);
     if (!UpdateResource(handle, MAKEINTRESOURCE(resourceType),
             MAKEINTRESOURCE(resourceId),
             MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL), resourceData,
@@ -320,10 +356,12 @@ static PyObject *ExtEndUpdateResource(
 {
     BOOL discardChanges;
     HANDLE handle;
+    PyObject *handle_obj;
 
     discardChanges = FALSE;
-    if (!PyArg_ParseTuple(args, "i|i", &handle, &discardChanges))
+    if (!PyArg_ParseTuple(args, "O|i", &handle_obj, &discardChanges))
         return NULL;
+    handle = PyLong_AsVoidPtr(handle_obj);
     if (!EndUpdateResource(handle, discardChanges)) {
         PyErr_SetExcFromWindowsErr(PyExc_WindowsError, GetLastError());
         return NULL;
@@ -371,9 +409,10 @@ static PyObject *ExtGetSystemDir(
     PyObject *self,                     // passthrough argument
     PyObject *args)                     // arguments (ignored)
 {
-    OLECHAR dir[MAX_PATH + 1];
-    if (GetSystemDirectoryW(dir, sizeof(dir)))
-        return PyUnicode_FromWideChar(dir, -1);
+    wchar_t dir_name[MAX_PATH + 1];
+
+    if (GetSystemDirectoryW(dir_name, sizeof(dir_name)))
+        return PyUnicode_FromWideChar(dir_name, -1);
     PyErr_SetExcFromWindowsErr(PyExc_RuntimeError, GetLastError());
     return NULL;
 }
@@ -387,28 +426,122 @@ static PyObject *ExtGetWindowsDir(
     PyObject *self,                     // passthrough argument
     PyObject *args)                     // arguments (ignored)
 {
-    OLECHAR dir[MAX_PATH + 1];
-    if (GetWindowsDirectoryW(dir, sizeof(dir)))
-        return PyUnicode_FromWideChar(dir, -1);
+    wchar_t dir_name[MAX_PATH + 1];
+
+    if (GetWindowsDirectoryW(dir_name, sizeof(dir_name)))
+        return PyUnicode_FromWideChar(dir_name, -1);
     PyErr_SetExcFromWindowsErr(PyExc_RuntimeError, GetLastError());
     return NULL;
 }
-#endif
+
+
+//-----------------------------------------------------------------------------
+// ExtUpdateCheckSum()
+//   Update the CheckSum into the specified executable.
+//-----------------------------------------------------------------------------
+static PyObject *ExtUpdateCheckSum(
+    PyObject *self,                     // passthrough argument
+    PyObject *args)                     // arguments
+{
+    PyObject *executable, *results;
+    wchar_t *filename;
+    HANDLE fhandle, fmap = NULL;
+    PVOID mmap = NULL;
+    DWORD filesize, header_sum, check_sum, last_error = 0;
+    PIMAGE_NT_HEADERS headers;
+    BOOL succeeded = TRUE;
+
+    if (!PyArg_ParseTuple(args, "U", &executable)) {
+        PyErr_Format(
+            PyExc_RuntimeError, "Invalid parameter."
+        );
+        return NULL;
+    }
+
+    filename = PyUnicode_AsWideCharString(executable, NULL);
+    if (!filename) {
+        return PyErr_NoMemory();
+    }
+
+    fhandle = CreateFileW((LPCWSTR) filename,
+                          GENERIC_READ | GENERIC_WRITE,
+                          FILE_SHARE_READ,
+                          NULL,
+                          OPEN_EXISTING,
+                          FILE_ATTRIBUTE_NORMAL,
+                          NULL);
+    if (fhandle == INVALID_HANDLE_VALUE) {
+        succeeded = FALSE;
+        last_error = GetLastError();
+    }
+    if (succeeded) {
+        filesize = GetFileSize(fhandle, NULL);
+        if (filesize == INVALID_FILE_SIZE) {
+            succeeded = FALSE;
+            last_error = GetLastError();
+        }
+    }
+    if (succeeded) {
+        fmap = CreateFileMapping(fhandle, NULL, PAGE_READWRITE, 0, 0, NULL);
+        if (!fmap) {
+            succeeded = FALSE;
+            last_error = GetLastError();
+        }
+    }
+    if (succeeded) {
+        mmap = MapViewOfFile(fmap, FILE_MAP_WRITE, 0, 0, 0);
+        if (mmap == NULL) {
+            succeeded = FALSE;
+            last_error = GetLastError();
+        }
+    }
+    if (succeeded) {
+        headers = CheckSumMappedFile(mmap, filesize, &header_sum, &check_sum);
+        if (headers == NULL) {
+            succeeded = FALSE;
+            last_error = GetLastError();
+        } else if (header_sum != 0 && header_sum != check_sum) {
+            headers->OptionalHeader.CheckSum = check_sum;
+        }
+    }
+
+    // clean up
+    if (mmap)
+        UnmapViewOfFile(mmap);
+    if (fmap)
+        CloseHandle(fmap);
+    if (fhandle)
+        CloseHandle(fhandle);
+    if (filename)
+        PyMem_Free(filename);
+    if (!succeeded) {
+        PyErr_SetExcFromWindowsErrWithFilenameObject(
+            PyExc_WindowsError, last_error, executable
+        );
+        return NULL;
+    }
+
+    // return a Tuple[int, int] (so we can check the values)
+    results = Py_BuildValue("(ii)", header_sum, check_sum);
+    if (results)
+        return results;
+    Py_INCREF(Py_None);
+    return Py_None;
+}
 
 
 //-----------------------------------------------------------------------------
 // Methods
 //-----------------------------------------------------------------------------
 static PyMethodDef g_ModuleMethods[] = {
-#ifdef MS_WINDOWS
+    { "AddIcon", ExtAddIcon, METH_VARARGS },
     { "BeginUpdateResource", ExtBeginUpdateResource, METH_VARARGS },
     { "UpdateResource", ExtUpdateResource, METH_VARARGS },
     { "EndUpdateResource", ExtEndUpdateResource, METH_VARARGS },
-    { "AddIcon", ExtAddIcon, METH_VARARGS },
     { "GetDependentFiles", ExtGetDependentFiles, METH_VARARGS },
     { "GetSystemDir", ExtGetSystemDir, METH_NOARGS },
     { "GetWindowsDir", ExtGetWindowsDir, METH_NOARGS },
-#endif
+    { "UpdateCheckSum", ExtUpdateCheckSum, METH_VARARGS },
     { NULL }
 };
 
@@ -439,13 +572,11 @@ PyMODINIT_FUNC PyInit_util(void)
     module = PyModule_Create(&g_ModuleDef);
     if (!module)
         return NULL;
-#ifdef MS_WINDOWS
     g_BindErrorException = PyErr_NewException("cx_Freeze.util.BindError",
             NULL, NULL);
     if (!g_BindErrorException)
         return NULL;
     if (PyModule_AddObject(module, "BindError", g_BindErrorException) < 0)
         return NULL;
-#endif
     return module;
 }
